/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "wmovie.hh"
#include "jsonStack.hh"

#include "Omega_c/wcondition.hh"
#include "Omega_c/wtransform.hh"
#include "Omega_c/wmeasure.hh"
#include "Omega_c/weventlist.hh"

#include "html/document.hh"
#include "html/image.hh"
#include "html/link.hh"
#include "html/table.hh"
#include "html/Table2C.hh"
#include "html/writer.hh"

#include <fstream>
#include <memory>
#include <sstream>
#include <stdexcept>
#include <unistd.h>

EXECDAT(wmovie)

using namespace wpipe;
using namespace std;

static const double dInf(1.0/0.0);

/** wmovie plots omega-transform plots in real time
 */
wmovie::wmovie(int argc, const char* argv[])
   : DatEnv(argc, argv), _glitch_threshold(18.5), _flush(0), _stride(64),
     _tWindow(32), _fSample(0), _maxglitch(10), _dtinflation(2.5),
     _bwinflation(2.5), _debug(0), _oscList(getDacc()), _max_frames(8),
     _current(0)
{
   string conf_file;
   for (int i=1; i<argc; i++) {
      string argi = argv[i];
      if (argi == "--config") {
	 conf_file = argv[++i];
      }
      else if (isDatEnvArg(argi.c_str())) {
	 i++;
      }
      else {
	 cerr << "wmovie: unrecognized command line option: " << argi << endl;
      }
   }

   //-----------------------------------  Read the configuration
   try {
      configure(conf_file);
   } catch (exception& e) {
      cerr << "caught exception in configuration: " << e.what() << endl;
      finish();
   }

   //-----------------------------------  Open the figure handle;
   _figureHandle.open();
   _figureHandle.set_palette("viridis");
   _figureHandle.set_size(28, 20);
}
  
//======================================  Destructor
wmovie::~wmovie(void) {
   //_figureHandle.close();
}

//======================================  Configuration
void 
wmovie::configure(std::string& file) {
   string condfile("osc_conditions.txt");
   _qRange = {3.32, 50.0};
   _fRange = {10.0, dInf};
   
   jsonStack js(file);
   
   if (!js.fetch_data("channel", _channel)) {
      throw runtime_error("wmovie: Channel not specified");
   }
   js.fetch_data("osc_file", condfile);
   js.fetch_data("osc_condition", _oscCondition);
   js.fetch_data("debug", _debug);
   double interval = 0;
   if (js.fetch_data("stride", interval)) {
      _stride = interval;
   }
   if (js.fetch_data("t-window", interval)) {
      _tWindow = interval;
   }
   js.fetch_data("f-minimum", _fRange.at(0));
   js.fetch_data("f-maximum", _fRange.at(1));
   js.fetch_data("sample-rate", _fSample);
   js.fetch_data("q-minimum", _qRange.at(0));
   js.fetch_data("q-maximum", _qRange.at(1));
   
   //------------------------------------  Set up the operating state conditions
   if (!_oscCondition.empty()) {
      _oscList.readConfig(condfile);
   }
   
   //------------------------------------  Set configuration
   getDacc().addChannel(_channel);
   getDacc().setStride(1.0);
   // getDacc().setStride(_tWindow*0.5);
   // setStrideAlignment(int(_tWindow*0.5), 0);
}

//======================================  Set up at the start of a run.
void 
wmovie::setup(double fSample) {
   const double maximumMismatch(0.2), highPassCutoff(10.0);
   const double lowPassCutoff(dInf), whiteningDuration(_flush);
   const double transientFactor(1.0);
   const int minAllowableIndependents(50);
   bool adjustLimits = false;
   //-----------------------------------  Define tile for Q-transform space
   _tiling.init(_stride, _qRange, _fRange, _fSample, maximumMismatch, 
		highPassCutoff, lowPassCutoff, whiteningDuration,
		transientFactor, minAllowableIndependents, adjustLimits,
		_debug);
   _data.clear();
   _data.push_back(TSeries());
}

//======================================  Process a stride.
void 
wmovie::ProcessData(void) {
   bool doubleWhiten(false);

   //------------------------------------  check the osc condition;
   if (_oscCondition.empty() || _oscList.satisfied(_oscCondition)) {
      TSeries* ts = getDacc().refData(_channel);
      if (!ts) throw runtime_error("wmovie: Data series is empty!");
      double sampleRate = 1.0 / double(ts->getTStep());
      if (_fSample == 0.0) _fSample = sampleRate;
      Time t0 = ts->getStartTime();
      if (t0 != _current) {
	 setup(_fSample);
      }
  
      //------------------------------------  resample data
      tser_vect raw_data;
      raw_data.push_back(*ts);
      tser_vect resdata = _resampler.wresample(raw_data, _fSample);
      if (_debug) {
	 cerr << "Resampled channel " << _channel << " start: " 
	      << resdata[0].getStartTime() << " end: "
	      << resdata[0].getEndTime() << endl;
      }
      if (_data[0].Append(resdata[0])) {
	 throw runtime_error("wmovie: Failed to append resampled series.");
      }
      // cout << "Accumulated data start: " << endl;
      // _data[0].dump_header(cout);
      if (_data[0].getInterval() >= _stride) {

	 //------------------------------------  Condition data and get dfts
	 //  Note that the standard Omega conditioner uses forward/backward
	 //  high-pass filtering so no history information is needed. It would
	 //  seem to me to be more effective to stream the data. This would
	 //  reduce the computation needed for the cnditioning step, especially
	 //  if the psd estimate csn be streamed effectively.
	 //
	 tser_vect condIn;
	 Time startTime = _data[0].getStartTime();
	 condIn.push_back(_data[0].extract(startTime, _stride));
	 wcondition conditioner(condIn, _tiling, doubleWhiten);

	 dft_vect condDft, coeffs;
	 conditioner.whitenedDFT(condDft);
	 conditioner.coefficientDFT(coeffs);

	 //---------------------------------  Transformation
	 const double outlierFactor = 1.0;
	 Time tRef = startTime + _stride * 0.5;
	 wtransform transform(condDft[0], _tiling, coeffs[0], outlierFactor,
			      _channel);

	 //-----------------------------  Get properties
	 dble_vect tLimits = {-_tWindow*0.5, _tWindow*0.5};
	 wmeasure properties(transform, _tiling, startTime, tRef, tLimits, 
			     _fRange, _qRange, _debug);
	 double    qBest  = properties[0].peak.Q;
	 double    normE  = properties[0].peak.NormalizedEnergy;
	 dble_vect qSelect = {qBest, qBest};
	 size_t p = _tiling.nearest_plane(qBest);
	 //properties.dump(cout);
	 
	 //---------------------------------  Get triggers
	 int maximumSignificants(10000);
	 double vetoThreshold(dInf), uncertaintyFactor(0.0);
	 double correlationFactor(0.0);
	 weventlist thlist, dslist;
	 thlist.wthreshold(transform[0], _tiling, _glitch_threshold, tRef,
			   tLimits, _fRange, qSelect, maximumSignificants,
			   uncertaintyFactor, _debug);
	 //thlist.display(cout);

	 // excluding overlapping significant tiles
	 dslist.wselect(thlist, _dtinflation, _bwinflation, 
			maximumSignificants, _debug);
	 
	 //dslist.display(cout);
	 size_t nGlitch = dslist.size();
	 
	 //-----------------------------  table0 (1, 2) contains everything
	 unique_ptr<html::table>  table0(new html::table());
	 table0->addColumn("");
	 table0->addColumn("");
	 table0->addRow();
      
	 //--------------------------------  Column 1 contains status (table1)
	 unique_ptr<html::Table2C> table1(new html::Table2C());
	 table1->addRow("CenterTime", tRef, "%s");
	 table1->addRow("Qval", qBest);
	 table1->addRow("Peak Energy", normE);
	 table1->addRow("Number of events", nGlitch);
	 table0->refCell(0, 0).addObject(table1.release());
	 table0->refCell(0, 0).lineBreak();

	 //---------------------------------  and loud event list
	 unique_ptr<html::table>  table2(new html::table());
	 table2->addColumn("#");
	 table2->addColumn("Frequency");
	 table2->addColumn("SNR");
	 table2->addColumn("Time");

	 for (size_t i=0; i<dslist.size() && i<_maxglitch; i++) {
	    table2->addRow();
	    table2->insertData(i, 0, long(i+1));
	    table2->insertData(i, 1, dslist[i].frequency, 6, 2);	    
	    table2->insertData(i, 2, sqrt(dslist[i].normalizedEnergy * 2 - 1),
			       6, 2);
	    table2->insertData(i, 3, dslist[i].t_offset, 6, 3);
	 }
	 table0->refCell(0,0).addObject(table2.release());

	 //-----------------------------  create image file
	 string outputDirectory("/home/jzweizig/public_html/wmovie/");
	 int plotHorizontalResolution(512), plotVerticalResolution(512);
	 dble_vect NormErgRange = {0.0, 25.0};
	 _figureHandle.clear();
	 _figureHandle.wspectrogram(transform[0], _tiling, tRef, tLimits, 
				    _fRange, qSelect, NormErgRange, 
				    plotHorizontalResolution,
				    plotVerticalResolution);

	 string channelName = _channel;
	 size_t inx = channelName.find(':');
	 while (inx != string::npos) {
	    channelName[inx] = '-';
	    inx = channelName.find(':');
	 }
	 // print whitened spectrogram to file
	 ostringstream figName;
	 figName << channelName << "_" << tRef.getS() << "-" << _tWindow;
	 string figBasePath = outputDirectory + figName.str();
	 _figureHandle.wprintfig(figBasePath);

	 //-----------------------------  Add the image to table0
	 table0->refCell(0, 1).addObject(new html::image(figName.str()+".png"));
	 
	 //-----------------------------  Build html document
	 html::document page;
	 page.addObject(table0.release());
	 page.setRefresh(_tWindow/2.0);

	 //-----------------------------  Add the movie vector
	 string thumb_width = "200";
	 unique_ptr<html::table> mvtable(new html::table());
	 for (size_t i=0; i<_max_frames; i++) {
	    if (i < _movie.size()) {
	       ostringstream title;
	       title << _movie[i]._gps_time.getS();
	       mvtable->addColumn(title.str());
	    } else {
	       mvtable->addColumn("");
	    }
	 }
	 mvtable->addRow();
	 for (size_t i=0; i<_movie.size(); i++) {
	    html::link imagelink;
	    imagelink.setAddr(_movie[i]._html_path);
	    html::object& l = imagelink.addObject(html::image(_movie[i]._thumb_path));
	    dynamic_cast<html::image&>(l).setWidth(thumb_width);
	    mvtable->refCell(0, i).addObject(imagelink);
	 }
	 page.lineBreak();
	 page.addObject(mvtable.release());
	 
	 //-----------------------------  Write out the html document
	 ostringstream docname;
	 docname << "wmovie-" << tRef.getS() << ".html";
	 ofstream docFile((outputDirectory + docname.str()).c_str());
	 html::writer docWriter(docFile);
	 page.write(docWriter);

	 //----------------------------- Point current to this
	 string cur_link  = outputDirectory + "current.html";
	 string prev_link = outputDirectory + "previous.html";
	 if (unlink(prev_link.c_str()) && errno != ENOENT) {
	    cerr << "wmovie: Unable to remove previous link." << endl;
	 }
	 if (rename(cur_link.c_str(), prev_link.c_str()) && errno != ENOENT) {
	    cerr << "wmovie: Unable to rename current link." << endl;
	 }	 
	 if (symlink(docname.str().c_str(), cur_link.c_str())) {
	    cerr << "wmovie: Unable to symlink current frame." << endl;
	 }

	 //-----------------------------  add a frame to the movie
	 movie_frame frame;
	 frame._gps_time   = tRef;
	 frame._html_path  = docname.str();
	 frame._thumb_path = figName.str() + ".thumb.png";
	 if (_movie.size() >= _max_frames) {
	    _movie.erase(_movie.begin(),
			 _movie.begin() + _max_frames +1 - _movie.size());
	 }
	 _movie.push_back(frame);
	 
	 //----------------------------- remove front data
	 _data[0].eraseStart(_tWindow * 0.5);
	 
      } // end 'if enough data'

      _current = ts->getEndTime();
   }
}
